/**
 * Copyright (C) cedarsoft GmbH.
 *
 * Licensed under the GNU General Public License version 3 (the "License")
 * with Classpath Exception; you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 *         http://www.cedarsoft.org/gpl3ce
 *         (GPL 3 with Classpath Exception)
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3 only, as
 * published by the Free Software Foundation. cedarsoft GmbH designates this
 * particular file as subject to the "Classpath" exception as provided
 * by cedarsoft GmbH in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 3 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 3 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact cedarsoft GmbH, 72810 Gomaringen, Germany,
 * or visit www.cedarsoft.com if you need additional information or
 * have any questions.
 */

package com.cedarsoft.gdao.async;

import com.cedarsoft.async.AsyncCallSupport;
import com.cedarsoft.async.CallbackCaller;
import com.cedarsoft.cache.Cache;
import com.cedarsoft.cache.HashedCache;
import com.cedarsoft.gdao.GenericDao;
import com.cedarsoft.gdao.GenericDaoManager;
import com.cedarsoft.gdao.LockProvider;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import java.lang.Override;
import java.util.Iterator;

/**
 * Special implementation of {@link GenericDaoManager} that
 * returns {@link AsynchronousDao}s.
 */
public final class AsynchronousDaoManager implements GenericDaoManager {

  @Nonnull
  private static final Log log = LogFactory.getLog( AsynchronousDaoManager.class );
  @Nonnull
  private final AsyncCallSupport<AsyncDaoCreator<?>> asyncCallSupport = new AsyncCallSupport<AsyncDaoCreator<?>>();

  @SuppressWarnings( {"MismatchedQueryAndUpdateOfCollection"} )
  @Nonnull
  private final Cache<Class<Object>, AsynchronousDao<Object>> daoCache = new HashedCache<Class<Object>, AsynchronousDao<Object>>(
      new Cache.Factory<Class<Object>, AsynchronousDao<Object>>() {
        @Override
        @Nonnull
        public AsynchronousDao<Object> create( @Nonnull Class<Object> key ) {
          return asyncCallSupport.invoke( new AsyncDaoCreator<Object>( key ) );
        }
      }
  );

  public AsynchronousDaoManager( @Nonnull final String description, @Nonnull final GenericDaoManager delegatingManager ) {
    asyncCallSupport.initializeWorker( new CallbackCaller<AsyncDaoCreator<?>>() {
      @Override
      public Object call( @Nonnull AsyncDaoCreator<?> callback ) throws Exception {
        log.debug( "executing callback" );
        return callback.execute( delegatingManager );
      }

      @Override
      @Nonnull
      public String getDescription() {
        return description;
      }
    } );
  }

  /**
   * Returns the asynchronous Dao
   *
   * @param type the type
   * @return the asynchronous dao for the given type
   */
  @Override
  @Nonnull
  public <T> AsynchronousDao<T> getDao( @Nonnull Class<T> type ) {
    return ( AsynchronousDao<T> ) daoCache.get( type );
  }

  @Override
  @Nonnull
  public <T> GenericDao<T> getDao( @Nonnull Class<T> type, @Nullable LockProvider<T> lockProvider ) {
    if ( lockProvider == null ) {
      return getDao( type );
    } else {
      throw new UnsupportedOperationException( "Lock provider not supported" );
    }
  }

  @Override
  public void shutdown() {
    for ( Iterator<AsynchronousDao<Object>> it = daoCache.values().iterator(); it.hasNext(); ) {
      AsynchronousDao<Object> dao = it.next();
      dao.shutdown();
      it.remove();
    }

    asyncCallSupport.shutdown();
  }

  private static class AsyncDaoCreator<T> {
    @Nonnull
    private final Class<T> key;

    private AsyncDaoCreator( @Nonnull Class<T> key ) {
      this.key = key;
    }

    @Nonnull
    public AsynchronousDao<T> execute( @Nonnull GenericDaoManager delegatingManager ) {
      log.debug( "Creating a new asynchronious dao for " + key.getName() );
      GenericDao<T> dao = delegatingManager.getDao( key );
      AsynchronousDao<T> asynchronousDao = new AsynchronousDao<T>();
      asynchronousDao.initializeDelegatingDao( dao );
      return asynchronousDao;
    }
  }
}
